from os import path
from unittest.mock import patch, MagicMock

import pytest

from . import commands


@patch('builtins.print')
def test_dupe_ingress(mock_print: MagicMock):
    env_file = path.join(path.dirname(__file__), 'e2e', 'dupe-ingress', 'environments.yaml')
    with pytest.raises(commands.CliError):
        commands.lint_environment(environments_file=env_file, env_name='test')
    assert mock_print.call_args_list[0][0][0].startswith('ERROR: Ingress')



import os
import subprocess
import sys
from typing import List, Dict

import logging
import ruamel.yaml
import ruamel.yaml.error
from blessings import Terminal
from ruamel.yaml import YAMLError

from .. import helm, updates, github
from .. import model
from ..inspection import validate_environment
from ..providers import get_cluster_provider
from ..thriftutils import SchemaError
from ..updates import find_updates_for_env, find_updates_for_releases

t = Terminal()
logger = logging.getLogger(__name__)


class CliError(Exception):
    pass


def lint_environment(environments_file: str, cluster: str=None, env_name: str=None, **kwargs):
    if not cluster and not env_name:
        raise CliError('please specify either --cluster or ENV')
    env_map = cluster_env_map(environments_file, cluster, env_name)
    errors = []
    for c, el in env_map.items():
        for env in el:
            errors.extend(validate_environment(env))
    for error in errors:
        print('{t.red}ERROR{t.normal}: {msg}'.format(msg=str(error), t=t), file=sys.stderr)

    if len(errors) > 0:
        raise CliError('validation failed')


def cluster_env_map(environments_file: str,
                    cluster: str=None,
                    env_name: str=None) -> Dict[model.Cluster, List[model.Environment]]:
    kcd_model = _load_model(environments_file)
    if cluster:
        try:
            return {kcd_model.get_cluster(cluster): kcd_model.environments_in_cluster(cluster)}
        except KeyError as e:
            raise CliError('no such cluster: {}'.format(e))
    try:
        e = kcd_model.get_environment(env_name)
        try:
            return {kcd_model.get_cluster(e.clusterName): [e]}
        except KeyError as e:
            raise CliError('no such cluster: {}'.format(e))
    except KeyError as e:
        raise CliError('no such environment: {}'.format(e))


def use_env_context(environments_file: str, env: str, **kwargs):
    kcd_model = _load_model(environments_file)
    try:
        kcd_model.get_environment(env)
    except KeyError as e:
        raise CliError('no such environment: {}'.format(e))
    run_command(helm.use_context_command(env))


def init_contexts(environments_file: str, cluster: str, env_name: str, contexts_only: bool=False, gitlab=False, dry_run=False, **kwargs):
    if not cluster and not env_name:
        raise CliError('please specify either --cluster or ENV')
    env_map = cluster_env_map(environments_file, cluster, env_name)
    commands_to_run = []
    commands_to_run.extend(model.config().init_commands())
    for c, el in env_map.items():
        cp = get_cluster_provider(c, gitlab)
        if not contexts_only and not gitlab:
            commands_to_run.extend(cp.cluster_init_commands())
        for e in el:
            commands_to_run.extend(cp.context_init_commands(e))
    run_commands(commands_to_run, dry_run=dry_run)


def run_command(cmd: List[str], dry_run=False):
    if dry_run:
        print(' '.join(cmd))
        return
    try:
        logger.debug('Executing: "%s"', ' '.join(cmd))
        return subprocess.run(cmd, check=True)
    except subprocess.CalledProcessError as e:
        raise CliError(e)


def run_commands(cmds: List[List[str]], dry_run=False):
    for cmd in cmds:
        run_command(cmd, dry_run)


def json2yaml(**kwargs):
    import json
    from ruamel import yaml
    obj = json.load(sys.stdin)
    yaml.safe_dump(obj, sys.stdout)


def diff_release(environments_file: str, releases: List[str], env: str, **kwargs):
    kcd_config = _load_model(environments_file)
    e = kcd_config.get_environment(env)
    for release in releases:
        r = e.named_release(release)
        cmd = helm.generate_helm_diff_argv(r, e, r.from_file)
        run_command(cmd)


def observe_new_image(environments_file: str, image: str, patch: bool, submit_pr: bool, **kwargs):
    kcd_config = _load_model(environments_file)
    image_repo, image_tag = image.split(':')
    image_index = kcd_config.image_index
    patched_files = []
    if image_repo in image_index:
        for release in image_index[image_repo]:
            update_list = updates.release_wants_tag_update(release, image_tag)
            for update in update_list:
                print('release: {release} tagValue: {tag}'.format(release=update.release.name, tag=update.tag_value))
            if len(update_list) > 0 and patch:
                patch_releases_file(release.from_file, update_list)
                patched_files.append(release.from_file)
    if submit_pr and len(patched_files) > 0:
        github.pr_from_files(patched_files, 'Observed image {image}'.format(image=image))


def apply_env(environments_file: str,
              init: bool=False,
              cluster: str=None,
              dry_run: bool=False,
              debug: bool=False,
              env_name: str=None,
              releases=None,
              **kwargs):
    commands_to_run = []
    for cluster, envs in cluster_env_map(environments_file, cluster, env_name).items():
        cp = get_cluster_provider(cluster)
        if init:
            commands_to_run.extend(cp.cluster_init_commands())
        for environment in envs:
            logger.info('Collecting commands for environment "%s"', environment.name)
            try:
                deploy_cmds = helm.deploy_commands(environment,
                                                   dry_run=dry_run,
                                                   limit_to_releases=releases,
                                                   debug=debug)
            except ValueError as e:
                raise CliError(str(e))
            if init and len(deploy_cmds) > 0 and init:
                commands_to_run.extend(cp.context_init_commands(environment))
            commands_to_run.extend(deploy_cmds)
    for cmd in commands_to_run:
        print('{t.yellow}{cmd}{t.normal}'.format(cmd=' '.join(cmd), t=t))
        run_command(cmd)


def dump_env(environments_file, cluster: str=None, env=None, **kwargs):
    target_envs = resolve_envs(env, cluster, file_name=environments_file)
    for cmd in model.config().init_commands():
        print('{t.yellow}{cmd}{t.normal}'.format(cmd=' '.join(cmd), t=t))
    for environment in target_envs:
        print('{t.green}Environment:{t.normal} {env_name}'.format(env_name=environment.name, t=t))
        for cmd in environment.init_commands():
            print('{t.yellow}{cmd}{t.normal}'.format(cmd=' '.join(cmd), t=t))
        for cmd in helm.deploy_commands(environment, dry_run=False):
            print(' '.join(cmd))
        print('')


def list_kind(environments_file, kind, **kwargs):
    kcd_model = _load_model(environments_file)
    if kind == 'env' or kind == 'environment':
        for environment in kcd_model:
            print(environment.name)
    elif kind == 'release' or kind == 'releases' or kind == 'rel':
        for environment in kcd_model:
            for release in environment.all_releases:
                print('{env} -r {release}'.format(env=environment.name, release=release.name))
    elif kind == 'cluster' or kind == 'clusters':
        for cluster in kcd_model.all_clusters():
            print('{cluster.name}'.format(cluster=cluster))
    else:
        raise CliError('unknown kind "{}"'.format(kind))


def poll_registries(environments_file, cluster=None, env=None, releases=None, patch=False, **kwargs):
    target_envs = resolve_envs(env, cluster, file_name=environments_file)
    for te in target_envs:
        if releases is None:
            logger.info('polling environment: "%s"', te.name)
            file_updates = find_updates_for_env(te)
        else:
            release_objs = [te.named_release(r) for r in releases]
            logger.info('polling env:%s releases: %s', te.name, ' '.join(releases))
            file_updates = find_updates_for_releases(release_objs, te)
        for release_file, update_list in file_updates.items():
            for update in update_list:
                print(
                    '{env}/{release}:\n\tfile: {file}\n\timage: {image}\n\ttag: {tag_from} -> {tag_to}'.format(
                        env=te.name,
                        file=release_file,
                        release=update.release.name,
                        image=update.image_repo,
                        tag_from=update.old_tag,
                        tag_to=update.new_tag,
                    ))
            if patch:
                patch_releases_file(release_file, update_list)


def patch_releases_file(releases_file: str, updates_list: List[updates.ImageUpdate]):
    logger.debug('loading releases file: "%s"', releases_file)
    mod_yaml = _load_yaml(releases_file)
    for update in updates_list:
        for mod_rel in mod_yaml['releases']:
            if mod_rel['name'] == update.release.name:
                if 'values' not in mod_rel:
                    mod_rel['values'] = []
                found_val = False
                for yv in mod_rel['values']:
                    if yv['key'] == update.tag_value:
                        logger.debug('patching value "{}"'.format(update.tag_value))
                        yv['value'] = update.new_tag
                        found_val = True
                        break
                if not found_val:
                    mod_rel['values'].append({'key': update.tag_value, 'value': update.new_tag})
    logger.debug('saving patched file: {file}'.format(file=releases_file))
    _save_yaml(mod_yaml, releases_file)


def indent_file(files, **kwargs):
    for filename in files:
        try:
            _save_yaml(_load_yaml(filename), filename)
        except YAMLError as e:
            raise CliError('invalid YAML file "{}": {}'.format(filename, str(e)))


def resolve_envs(env_name: str, cluster: str, file_name: str) -> List[model.Environment]:
    if not env_name and not cluster:
        raise CliError('please specify either --cluster or an environment name')
    kcd_model = _load_model(file_name)
    if env_name is not None:
        try:
            return [kcd_model.get_environment(env_name)]
        except KeyError:
            raise CliError('no such environment: {}'.format(env_name))
    if kcd_model.has_cluster(cluster):
        return kcd_model.environments_in_cluster(cluster)
    raise CliError('no such cluster: {}'.format(cluster))


def _load_model(file_name: str):
    try:
        logger.debug('loading model from file %s', file_name)
        return model.load(file_name)
    except FileNotFoundError as e:
        raise CliError('file not found: {}'.format(e.filename))
    except SchemaError as e:
        raise CliError(e)


def _load_yaml(file_name: str):
    yaml = ruamel.yaml.YAML(typ='rt')
    yaml.indent(sequence=4, offset=2)
    yaml.preserve_quotes = True
    with open(file_name, 'r') as rf:
        return yaml.load(rf)


def _save_yaml(mod_yaml, file_name: str):
    yaml = ruamel.yaml.YAML(typ='rt')
    yaml.indent(sequence=4, offset=2)
    yaml.preserve_quotes = True
    tmp_file = file_name + '.new'
    with open(tmp_file, 'w') as wf:
        yaml.dump(mod_yaml, wf)
    os.rename(tmp_file, file_name)



import abc
from typing import List

import json

from .utils import kube_context
from .gen_py.ttypes import Environment, Cluster
from .gen_py import ttypes


class BaseClusterProvider(metaclass=abc.ABCMeta):
    def __init__(self, cluster: ttypes.Cluster):
        self.cluster = cluster

    @abc.abstractmethod
    def cluster_init_commands(self) -> List[List[str]]:
        pass

    @abc.abstractmethod
    def cluster_name(self) -> str:
        pass

    @abc.abstractmethod
    def user_name(self) -> str:
        pass

    def namespace(self, env: Environment) -> str:
        return env.kubeNamespace

    def context_init_commands(self, env: Environment) -> List[List[str]]:
        context_name = kube_context(env.name)
        return [[
            'kubectl', 'config', 'set-context', context_name,
            '--cluster', self.cluster_name(),
            '--user', self.user_name(),
            '--namespace', self.namespace(env)
        ]]


class GitLabClusterProvider(BaseClusterProvider):
    def cluster_init_commands(self) -> List[List[str]]:
        return []

    def cluster_name(self) -> str:
        return 'gitlab-deploy'

    def user_name(self) -> str:
        return 'gitlab-deploy'


class GkeClusterProvider(BaseClusterProvider):
    def cluster_init_commands(self) -> List[List[str]]:
        return [[
            'gcloud', 'container', 'clusters', 'get-credentials',
            '--project', self.cluster.provider.gke.project,
            '--zone', self.cluster.provider.gke.zone,
            self.cluster.provider.gke.clusterName
        ]]

    def cluster_name(self) -> str:
        return 'gke_{gke.project}_{gke.zone}_{gke.clusterName}'.format(gke=self.cluster.provider.gke)

    def user_name(self) -> str:
        return self.cluster_name()


class AksClusterProvider(BaseClusterProvider):
    def cluster_init_commands(self) -> List[List[str]]:
        return [[
            'az', 'aks', 'get-credentials',
            '--resource-group', self.cluster.provider.aks.resourceGroup,
            '--name', self.cluster.provider.aks.clusterName
        ]]

    def cluster_name(self) -> str:
        return self.cluster.provider.aks.clusterName

    def user_name(self) -> str:
        return 'clusterUser_{aks.resourceGroup}_{aks.clusterName}'.format(aks=self.cluster.provider.aks)


class MinikubeClusterProvider(BaseClusterProvider):
    def cluster_init_commands(self) -> List[List[str]]:
        return []

    def cluster_name(self) -> str:
        return 'minikube'

    def user_name(self) -> str:
        return 'minikube'


class DockerForDesktopProvider(BaseClusterProvider):
    def cluster_init_commands(self) -> List[List[str]]:
        return [
            ['kubectl', 'config', 'set-cluster', 'docker-for-desktop-cluster',
             '--insecure-skip-tls-verify=true', '--server=https://localhost:6443']
        ]

    def cluster_name(self) -> str:
        return 'docker-for-desktop-cluster'

    def user_name(self) -> str:
        return 'docker-for-desktop'


class ExistingContextProvider(BaseClusterProvider):
    def __init__(self, cluster: ttypes.Cluster):
        import subprocess
        super().__init__(cluster)
        kube_config = json.loads(subprocess.check_output(['kubectl', 'config', 'view', '-o', 'json']).decode('utf-8'))
        self.kube_context = None
        name = cluster.provider.existingContext.contextName
        for context in kube_config['contexts']:
            if name in context['name'] and 'context' in context:
                self.kube_context = context['context']
        if self.kube_context is None:
            raise ValueError('context "{}" not found in existing kube config'.format(name))

    def cluster_init_commands(self) -> List[List[str]]:
        return []

    def cluster_name(self) -> str:
        return self.kube_context['cluster']

    def user_name(self) -> str:
        return self.kube_context['user']

    def namespace(self, env: Environment) -> str:
        return self.kube_context['namespace'] if 'namespace' in self.kube_context else 'default'


def get_cluster_provider(cluster: Cluster, gitlab_mode: bool=False) -> BaseClusterProvider:
    if gitlab_mode:
        return GitLabClusterProvider(cluster)
    if cluster.provider.gke:
        return GkeClusterProvider(cluster)
    elif cluster.provider.minikube:
        return MinikubeClusterProvider(cluster)
    elif cluster.provider.aks:
        return AksClusterProvider(cluster)
    elif cluster.provider.dockerForDesktop:
        return DockerForDesktopProvider(cluster)
    elif cluster.provider.existingContext:
        return ExistingContextProvider(cluster)
    raise ValueError('missing or unknown cluster provider')


def cluster_init_commands(cluster: Cluster, env: Environment) -> List[List[str]]:
    """
    :param cluster:  cluster object for the environment
    :param env:      which environment we should generate init command[s] for
    :return: a list of commands, each command is an argv list
    """
    cluster_provider = get_cluster_provider(cluster)
    commands = cluster_provider.cluster_init_commands()
    commands.extend(cluster_provider.context_init_commands(env))
    return commands



import logging
import subprocess
from copy import copy
from os import path
from tempfile import TemporaryDirectory
from typing import List, TextIO

from ruamel.yaml import YAML

from .helm import generate_helm_template_argv
from .model import Environment, Release
from .utils import resolve_file_path


class ValidationError(Exception):
    pass


def load_yaml(data: TextIO):
    yaml = YAML(typ='safe')
    return yaml.load_all(data)


def run_command(cmd: List[str], dry_run=False):
    logging.debug('Executing: "%s"', ' '.join(cmd))
    return subprocess.run(cmd, check=True)


def expand_release(rel: Release, env: Environment):
    if rel.resourceFiles:
        return expand_resource_files(rel.resourceFiles, rel.from_file)
    if rel.chart:
        if rel.chart.dir:
            return expand_helm_chart_dir(rel, env)
        elif rel.chart.reference:
            with TemporaryDirectory() as tmpdir:
                rel.chart = copy(rel.chart)
                rel.chart.dir = path.join(tmpdir, path.basename(rel.chart.reference))
                subprocess.run([
                    'helm', 'fetch', '--untar', '--untardir', tmpdir,
                    rel.chart.reference, '--version', rel.chart.version
                ])
                rel.chart.reference = None
                rel.chart.version = None
                return expand_helm_chart_dir(rel, env)


def expand_helm_chart_dir(rel: Release, env: Environment) -> List[dict]:
    cmd = generate_helm_template_argv(rel, env, rel.from_file)
    logging.debug('Running %r', cmd)
    output = subprocess.check_output(cmd).decode('utf-8')
    resources = load_yaml(output)
    if isinstance(resources, dict):
        resources = [resources]
    return resources


def expand_resource_files(file_names: List[str], from_file: str) -> List[dict]:
    resources = []
    for file_name in file_names:
        file_name = resolve_file_path(file_name, relative_to_file=from_file)
        with open(file_name, 'r') as f:
            tmp = load_yaml(f)
            if isinstance(tmp, dict):
                resources.append(tmp)
            else:
                resources.extend(tmp)
    return resources


def index_resource_list(resources: List[dict]):
    index = {}
    for obj in resources:
        key = '{kind}.{name}'.format(kind=obj['kind'], name=obj['metadata']['name'])
        index[key] = obj
    return index


def find_ingress_conflicts(resources: List[dict]) -> List[ValidationError]:
    errors = []
    path_owner = {}
    for obj in resources:
        if obj is None:
            continue
        logging.debug('checking obj %r', obj)
        if obj['kind'] == 'Ingress':
            ingress_name = obj['metadata']['name']
            for rule in obj['spec']['rules']:
                for http_path in rule['http']['paths']:
                    host_path = rule['host'] + (http_path['path'] if 'path' in http_path else '/')
                    if host_path in path_owner:
                        errors.append(ValidationError('Ingress host+path "{host_path}" defined in both {a} and {b}'.format(
                            host_path=host_path, a=path_owner[host_path], b=ingress_name
                        )))
                        continue
                    path_owner[host_path] = ingress_name
    return errors


def validate_environment(env: Environment) -> List[ValidationError]:
    resources = []
    errors = []
    for release in env.all_releases:
        logging.info('expanding release: %s', release.name)
        resources.extend(expand_release(release, env))
    logging.info('checking for: ingress conflicts')
    errors.extend(find_ingress_conflicts(resources))
    return errors



import argparse
import os
import sys

import argcomplete
import logging
from argcomplete import FilesCompleter
from blessings import Terminal

from .. import __version__
from .commands import (
    apply_env,
    dump_env,
    indent_file,
    init_contexts,
    json2yaml,
    lint_environment,
    list_kind,
    observe_new_image,
    poll_registries,
    use_env_context,
    CliError,
)

t = Terminal()
logger = logging.getLogger(__name__)


def parser(prog='kcd') -> argparse.ArgumentParser:
    p = argparse.ArgumentParser(prog=prog)
    # not used yet, but left here as a reminder to not steal the -c flag
    # p.add_argument('-c', '--config-file', help='path to configuration file')
    p.add_argument(
        '-f', '--environments-file',
        help='KubeCD config file file (default $KUBECD_ENVIRONMENTS or "environments.yaml")',
        metavar='FILE',
        default=os.getenv('KUBECD_ENVIRONMENTS', 'environments.yaml')).completer = FilesCompleter(
            allowednames=('.yaml', '.yml'), directories=False)
    p.add_argument(
        '--version',
        help='Show version and exit.',
        action='version',
        version='kubecd ' + __version__)
    p.add_argument('--verbose', '-v', help='Increase verbosity level', action='count', default=0)

    s = p.add_subparsers(dest='command', title='Subcommands', description='Use one of these sub-commands:')

    apply = s.add_parser('apply',
                         help='apply changes to Kubernetes')
    apply.add_argument('--dry-run', '-n', action='store_true', default=False,
                       help='dry run mode, only print commands')
    apply.add_argument('--debug', action='store_true', default=False,
                       help='run helm with --debug')
    apply.add_argument('--releases', '-r', action='append',
                       help='apply only these releases')
    apply.add_argument('--cluster', '-c', nargs='?', metavar='CLUSTER',
                       help='apply all environments in CLUSTER')
    apply.add_argument('--init', action='store_true', default=False,
                       help='Initialize credentials and contexts')
    apply.add_argument('env_name', nargs='?', metavar='ENV',
                       help='name of environment to apply, must be specified unless --cluster is')
    apply.set_defaults(func=apply_env)

    # diff = s.add_parser('diff', help='show diffs between running and git release')
    # diff.add_argument('--releases', '-r', help='which releases to diff', action='append')
    # diff.add_argument('env', nargs='?', help='name of environment')
    # diff.set_defaults(func=diff_release)

    poll_p = s.add_parser('poll',
                          help='poll for new images in registries')
    poll_p.add_argument('--patch', '-p', action='store_true',
                        help='patch releases.yaml files with updated version')
    poll_p.add_argument('--releases', '-r', action='append',
                        help='poll this specific release')
    poll_p.add_argument('--image', '-i',
                        help='poll releases using this image')
    poll_p.add_argument('--cluster', '-c',
                        help='poll all releases in this cluster')
    poll_p.add_argument('env', nargs='?',
                        help='name of environment to poll')
    poll_p.set_defaults(func=poll_registries)

    dump_p = s.add_parser('dump',
                          help='dump commands for one or all environments')
    dump_p.add_argument('env', nargs='?',
                        help='name of environment to dump')
    dump_p.set_defaults(func=dump_env)

    list_p = s.add_parser('list',
                          help='list clusters, environments or releases')
    list_p.add_argument('kind', choices=['env', 'release', 'cluster'],
                        help='what to list')
    list_p.set_defaults(func=list_kind)

    indent_p = s.add_parser('indent',
                            help='canonically indent YAML files')
    indent_p.add_argument('files', nargs='+',
                          help='file[s] to indent')
    indent_p.set_defaults(func=indent_file)

    observe = s.add_parser('observe',
                           help='observe a new image version')
    observe.add_argument('--image', '-i', metavar='IMAGE:TAG',
                         help='the image, including tag')
    observe.add_argument('--patch', action='store_true', default=False,
                         help='patch release files with updated tags')
    observe.add_argument('--submit-pr', action='store_true', default=False,
                         help='submit a pull request with the updated tags')
    observe.set_defaults(func=observe_new_image)

    completion_p = s.add_parser('completion',
                                help='print shell completion script')
    completion_p.set_defaults(func=print_completion, prog=prog)

    j2y = s.add_parser('json2yaml', help='JSON to YAML conversion utility (stdin/stdout)')
    j2y.set_defaults(func=json2yaml)

    init = s.add_parser('init', help='Initialize credentials and contexts')
    init.add_argument('--cluster',
                      help='Initialize contexts for all environments in a cluster')
    init.add_argument('--dry-run', '-n', action='store_true',
                      help='print commands instead of running them')
    init.add_argument('env_name', metavar='ENV', nargs='?',
                      help='environment to initialize')
    init.add_argument('--contexts-only', action='store_true',
                      help='initialize contexts only, assuming that cluster credentials are set up')
    init.add_argument('--gitlab', action='store_true',
                      help='grab kube config from GitLab environment')
    init.set_defaults(func=init_contexts)

    use = s.add_parser('use',
                       help='switch kube context to the specified environment')
    use.add_argument('env', metavar='ENV',
                     help='environment name')
    use.set_defaults(func=use_env_context)

    lint = s.add_parser('lint',
                        help='inspect the contents of a release, exits with non-0 if there are issues')
    lint.add_argument('--cluster',
                      help='Lint all environments in a cluster')
    lint.add_argument('env_name', metavar='ENV', nargs='?',
                      help='environment name')
    lint.set_defaults(func=lint_environment)

    return p


def verbose_log_level(v):
    if v == 0:
        return logging.WARNING
    if v == 1:
        return logging.INFO
    return logging.DEBUG


def print_completion(prog, **kwargs):
    shell = os.path.basename(os.getenv('SHELL'))
    if shell == 'bash' or shell == 'tcsh':
        sys.stdout.write(argcomplete.shellcode(prog, shell=shell))


def main():
    p = parser()
    argcomplete.autocomplete(p)
    args = p.parse_args()
    kwargs = args.__dict__
    if 'func' not in kwargs:
        p.print_help(sys.stderr)
        sys.exit(1)
    func = kwargs['func']
    del (kwargs['func'])
    logging.basicConfig(stream=sys.stderr,
                        format='{levelname} {message}',
                        style='{',
                        level=verbose_log_level(args.verbose))
    try:
        func(**kwargs)
    except CliError as e:
        print('{t.red}ERROR{t.normal}: {msg}'.format(msg=str(e), t=t), file=sys.stderr)


if __name__ == '__main__':
    main()
